(*
(C) Copyright Andreas Viktor Hess, DTU, 2015-2018

All Rights Reserved.

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are
met:

- Redistributions of source code must retain the above copyright
  notice, this list of conditions and the following disclaimer.

- Redistributions in binary form must reproduce the above copyright
  notice, this list of conditions and the following disclaimer in the
  documentation and/or other materials provided with the distribution.

- Neither the name of the copyright holder nor the names of its
  contributors may be used to endorse or promote products
  derived from this software without specific prior written
  permission.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
*)

(*  Title:      TLSExample.thy
    Author:     Andreas Viktor Hess, DTU
*)

section {* Theory: Proving Type-Flaw Resistance of the TLS Handshake Protocol *}
theory Example_TLS
imports Typed_Model
begin

subsection {* TLS example: Datatypes and functions setup *}
datatype ex_atom = PrivKey | SymKey | PubConst | Agent | Nonce | Bot

datatype ex_fun =
  clientHello | clientKeyExchange | clientFinished
| serverHello | serverCert | serverHelloDone
| finished | changeCipher | x509 | prfun | master | pmsForm
| sign | hash | crypt | pub | concat | privkey nat
| pubconst ex_atom nat

type_synonym ex_type = "(ex_fun, ex_atom) term_type"
type_synonym ex_var = "ex_type \<times> nat"

instance ex_atom::finite
proof
  let ?S = "UNIV::ex_atom set"
  have "?S = {PrivKey, SymKey, PubConst, Agent, Nonce, Bot}" by (auto intro: ex_atom.exhaust)
  thus "finite ?S" by (metis finite.emptyI finite.insertI) 
qed

type_synonym ex_term = "(ex_fun, ex_var) term"
type_synonym ex_terms = "(ex_fun, ex_var) terms"

primrec arity::"ex_fun \<Rightarrow> nat" where
  "arity changeCipher = 0"
| "arity clientFinished = 4"
| "arity clientHello = 5"
| "arity clientKeyExchange = 1"
| "arity concat = 5"
| "arity crypt = 2"
| "arity finished = 1"
| "arity hash = 1"
| "arity master = 3"
| "arity pmsForm = 1"
| "arity prfun = 1"
| "arity (privkey _) = 0"
| "arity pub = 1"
| "arity (pubconst _ _) = 0"
| "arity serverCert = 1"
| "arity serverHello = 5"
| "arity serverHelloDone = 0"
| "arity sign = 2"
| "arity x509 = 2"

fun public::"ex_fun \<Rightarrow> bool" where "public f = (case f of (privkey _) \<Rightarrow> False | _ \<Rightarrow> True)"

fun Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t::"ex_term list \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t [Fun pub [k],m] = ({k}, {m})"
| "Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t _ = ({}, {})"

fun Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n::"ex_term list \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n [k,m] = ({}, {m})"
| "Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n _ = ({}, {})"

fun Ana::"ex_term \<Rightarrow> (ex_terms \<times> ex_terms)" where
  "Ana (Fun crypt T) = Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t T"
| "Ana (Fun finished T) = ({}, set T)"
| "Ana (Fun master T) = ({}, set T)"
| "Ana (Fun pmsForm T) = ({}, set T)"
| "Ana (Fun serverCert T) = ({}, set T)"
| "Ana (Fun serverHello T) = ({}, set T)"
| "Ana (Fun serverHelloDone T) = ({}, set T)"
| "Ana (Fun sign T) = Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n T"
| "Ana (Fun x509 T) = ({}, set T)"
| "Ana _ = ({}, {})"


subsection {* TLS example: Locale interpretation *}
lemma assm1:
  "Ana t = (K,M) \<Longrightarrow> fv\<^sub>s\<^sub>e\<^sub>t K \<subseteq> fv t"
  "Ana t = (K,M) \<Longrightarrow> finite K"
  "Ana t = (K,M) \<Longrightarrow> (\<And>g S'. Fun g S' \<sqsubseteq> t \<Longrightarrow> length S' = arity g)
                \<Longrightarrow> k \<in> K \<Longrightarrow> Fun f T' \<sqsubseteq> k \<Longrightarrow> length T' = arity f"
  "Ana t = (K,M) \<Longrightarrow> K \<noteq> {} \<or> M \<noteq> {} \<Longrightarrow> Ana (t \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>, M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
by (rule Ana.cases[of "t"], auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)+

lemma assm2: "Ana (Fun f T) = (K, M) \<Longrightarrow> M \<subseteq> set T"
by (rule Ana.cases[of "Fun f T"]) (auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)

lemma assm5: "infinite {c. public c \<and> arity c = 0}"
proof -
  let ?\<tau> = "PubConst"
  let ?T = "(range (pubconst ?\<tau>))::ex_fun set"
  have *:
       "\<And>x y::nat. x \<in> UNIV \<Longrightarrow> y \<in> UNIV \<Longrightarrow> (pubconst ?\<tau> x = pubconst ?\<tau> y) = (x = y)"
       "\<And>x::nat. x \<in> UNIV \<Longrightarrow> pubconst ?\<tau> x \<in> ?T"
       "\<And>y::ex_fun. y \<in> ?T \<Longrightarrow> \<exists>x \<in> UNIV. y = pubconst ?\<tau> x"
    by auto
  have "?T \<subseteq> {c. public c \<and> arity c = 0}" by auto
  moreover have "\<exists>f::nat \<Rightarrow> ex_fun. bij_betw f UNIV ?T"
    using bij_betwI'[OF *] by blast
  hence "infinite ?T" by (metis nat_not_finite bij_betw_finite)
  ultimately show ?thesis using infinite_super by blast
qed

lemma assm6: "0 < arity f \<Longrightarrow> public f" by (cases f) simp_all

interpretation intruder_model arity public Ana
apply standard
by (metis assm1(1), metis assm1(2), metis assm1(3), rule Ana.simps,
    metis assm2, metis assm1(4), metis assm5)


subsection {* TLS Example: Typing function *}
definition \<Gamma>\<^sub>v::"ex_var \<Rightarrow> ex_type" where
  "\<Gamma>\<^sub>v v = (if (\<forall>t \<in> subterms (fst v). case t of
                (TComp f T) \<Rightarrow> arity f > 0 \<and> arity f = length T
              | _ \<Rightarrow> True)
           then fst v else TAtom Bot)"

fun \<Gamma>::"ex_term \<Rightarrow> ex_type" where
  "\<Gamma> (Var v) = \<Gamma>\<^sub>v v"
| "\<Gamma> (Fun (privkey _) _) = TAtom PrivKey"
| "\<Gamma> (Fun changeCipher _) = TAtom PubConst"
| "\<Gamma> (Fun serverHelloDone _) = TAtom PubConst"
| "\<Gamma> (Fun (pubconst \<tau> _) _) = TAtom \<tau>"
| "\<Gamma> (Fun f T) = TComp f (map \<Gamma> T)"


subsection {* TLS Example: Locale interpretation (typed model) *}
lemma assm7: "arity c = 0 \<Longrightarrow> \<exists>a. \<forall>X. \<Gamma> (Fun c X) = TAtom a" by (cases c) simp_all

lemma assm8: "0 < arity f \<Longrightarrow> \<Gamma> (Fun f X) = TComp f (map \<Gamma> X)" by (cases f) simp_all

lemma assm9: "infinite {c. \<Gamma> (Fun c []) = TAtom a \<and> public c}"
proof -
  let ?T = "(range (pubconst a))::ex_fun set"
  have *:
      "\<And>x y::nat. x \<in> UNIV \<Longrightarrow> y \<in> UNIV \<Longrightarrow> (pubconst a x = pubconst a y) = (x = y)"
      "\<And>x::nat. x \<in> UNIV \<Longrightarrow> pubconst a x \<in> ?T"
      "\<And>y::ex_fun. y \<in> ?T \<Longrightarrow> \<exists>x \<in> UNIV. y = pubconst a x"
    by auto
  have "?T \<subseteq> {c. \<Gamma> (Fun c []) = TAtom a \<and> public c}" by auto
  moreover have "\<exists>f::nat \<Rightarrow> ex_fun. bij_betw f UNIV ?T"
    using bij_betwI'[OF *] by blast
  hence "infinite ?T" by (metis nat_not_finite bij_betw_finite)
  ultimately show ?thesis using infinite_super by blast
qed

lemma assm10: "TComp f T \<sqsubseteq> \<Gamma> t \<Longrightarrow> arity f > 0"
proof (induction rule: \<Gamma>.induct)
  case (1 x)
  hence *: "TComp f T \<sqsubseteq> \<Gamma>\<^sub>v x" by simp
  hence "\<Gamma>\<^sub>v x \<noteq> TAtom Bot" unfolding \<Gamma>\<^sub>v_def by force
  hence "\<forall>t \<in> subterms (fst x). case t of
                (TComp f T) \<Rightarrow> arity f > 0 \<and> arity f = length T
              | _ \<Rightarrow> True"
    unfolding \<Gamma>\<^sub>v_def by argo
  thus ?case using * unfolding \<Gamma>\<^sub>v_def subtermeq_def by fastforce 
qed auto

lemma assm11: "wf\<^sub>t\<^sub>r\<^sub>m (\<Gamma> (Var x))"
proof -
  have "wf\<^sub>t\<^sub>r\<^sub>m (\<Gamma>\<^sub>v x)" unfolding \<Gamma>\<^sub>v_def wf\<^sub>t\<^sub>r\<^sub>m_def by auto 
  thus ?thesis by simp
qed

interpretation typed_model arity public Ana \<Gamma>
by standard (metis assm7, metis assm8, metis assm9, metis assm10, metis assm11, metis assm6)


subsection {* Automatically checking type-flaw resistance *}
definition is_subterm_closed where
  "is_subterm_closed M \<equiv> \<Union>(subterms ` set M) \<subseteq> set M"

definition is_Ana_closed where
  "is_Ana_closed M \<equiv> \<forall>s \<in> set M. case Ana s of (K,T) \<Rightarrow> K \<subseteq> set M"

definition is_TComp_var_instance_closed where
  "is_TComp_var_instance_closed M \<equiv> \<forall>v \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M).
     case \<Gamma>\<^sub>v v of
       (TComp f T) \<Rightarrow> (\<exists>t \<in> set M.
         case t of
           (Fun g U) \<Rightarrow> (\<Gamma> (Fun g U) = \<Gamma>\<^sub>v v \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U)
         | _ \<Rightarrow> False)
     | _ \<Rightarrow> True"

definition is_pairwise_var_disjoint where
  "is_pairwise_var_disjoint M \<equiv> \<forall>s \<in> set M. \<forall>t \<in> set M. s \<noteq> t \<longrightarrow> fv s \<inter> fv t = {}"

definition wf\<^sub>t\<^sub>r\<^sub>m_comp::"ex_term \<Rightarrow> bool" where
  "wf\<^sub>t\<^sub>r\<^sub>m_comp t \<equiv> (\<forall>t \<in> subterms t. case t of
                      (Fun f T) \<Rightarrow> arity f = length T
                    | _ \<Rightarrow> True)"

lemma wf\<^sub>t\<^sub>r\<^sub>m_code[code]: "wf\<^sub>t\<^sub>r\<^sub>m t = wf\<^sub>t\<^sub>r\<^sub>m_comp t"
proof
  show "wf\<^sub>t\<^sub>r\<^sub>m_comp t \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m t" unfolding wf\<^sub>t\<^sub>r\<^sub>m_comp_def wf\<^sub>t\<^sub>r\<^sub>m_def subtermeq_def by auto
  
  { fix s assume s: "s \<in> subterms t" "wf\<^sub>t\<^sub>r\<^sub>m t"
    hence "case s of Var x \<Rightarrow> True | Fun f T \<Rightarrow> arity f = length T" 
      unfolding wf\<^sub>t\<^sub>r\<^sub>m_comp_def wf\<^sub>t\<^sub>r\<^sub>m_def subtermeq_def by (cases s) auto
  } 
  thus "wf\<^sub>t\<^sub>r\<^sub>m t \<Longrightarrow> wf\<^sub>t\<^sub>r\<^sub>m_comp t" by (metis wf\<^sub>t\<^sub>r\<^sub>m_comp_def)
qed

definition is_wellformed_terms where
  "is_wellformed_terms M \<equiv> \<forall>t \<in> set M. wf\<^sub>t\<^sub>r\<^sub>m_comp t"

definition "max_var t \<equiv> Max (insert 0 (snd ` fv t))"

definition "var_rename n v \<equiv> Var (fst v, snd v + Suc n)"
definition "var_rename_inv n v \<equiv> Var (fst v, snd v - Suc n)"

fun var_disj' where
  "var_disj' _ [] = []"
| "var_disj' n (t#T) = t \<cdot> var_rename n#var_disj' (max_var t + Suc n) T"

definition "var_disj M \<equiv> var_disj' (Max (insert 0 (snd ` fv\<^sub>s\<^sub>e\<^sub>t (set M)))) M"

definition comp_tfr where
  "comp_tfr M \<equiv>
    is_wellformed_terms M \<and>
    is_subterm_closed M \<and>
    is_Ana_closed M \<and>
    is_TComp_var_instance_closed M \<and>
    is_pairwise_var_disjoint (var_disj M) \<and>
    (\<forall>s \<in> set (var_disj M). \<forall>t \<in> set (var_disj M).
      is_Fun s \<and> is_Fun t \<and> \<Gamma> s \<noteq> \<Gamma> t \<longrightarrow> mgu s t = None)"


subsection {* Lemma: comp_tfr implies tfr set *}
lemma var_rename_inv_comp: "t \<cdot> (var_rename n \<circ>\<^sub>s var_rename_inv n) = t"
proof (induction t)
  case (Fun f T)
  hence "map (\<lambda>t. t \<cdot> var_rename n \<circ>\<^sub>s var_rename_inv n) T = T" by (simp add: map_idI) 
  thus ?case by (metis subst_apply_term.simps(2)) 
qed (simp add: var_rename_def var_rename_inv_def)

lemma var_rename_fv_disjoint:
  "fv s \<inter> fv (t \<cdot> var_rename (max_var s)) = {}"
proof -
  have 1: "\<forall>v \<in> fv s. snd v \<le> max_var s" unfolding max_var_def by simp
  have 2: "\<And>n. \<forall>v \<in> fv (t \<cdot> var_rename n). snd v > n" unfolding var_rename_def by (induct t) auto
  show ?thesis using 1 2 by force
qed

lemma var_rename_is_renaming:
  "subst_range (var_rename n) \<subseteq> range Var"
  "subst_range (var_rename_inv n) \<subseteq> range Var"
unfolding var_rename_def var_rename_inv_def by auto

lemma var_rename_wt: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (var_rename n)" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (var_rename_inv n)"
by (auto simp add: var_rename_def var_rename_inv_def wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def \<Gamma>\<^sub>v_def)

lemma var_disj_is_var_renaming:
  assumes "t \<in> set M"
  shows "\<exists>m \<in> set (var_disj M). \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> subst_range \<delta> \<subseteq> range Var \<and> t = m \<cdot> \<delta>"
  using assms
proof -
  { fix n
    have "\<exists>m \<in> set (var_disj' n M). \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> subst_range \<delta> \<subseteq> range Var \<and> t = m \<cdot> \<delta>"
      using assms
    proof (induction M arbitrary: n)
      case (Cons m M)
      show ?case
      proof (cases "t \<in> set M")
        case False
        hence "t = m" using Cons.prems by simp
        hence "t = m \<cdot> var_rename n \<cdot> var_rename_inv n" by (metis var_rename_inv_comp subst_comp)
        moreover have "m \<cdot> var_rename n \<in> set (var_disj' n (m#M))" by simp
        ultimately show ?thesis by (metis var_rename_is_renaming(2) var_rename_wt(2))
      qed (simp add: Cons.IH del: subst_range.simps)
    qed simp
  } thus ?thesis by (metis var_disj_def)
qed

lemma var_disj_is_var_renaming':
  assumes "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "s = m \<cdot> \<delta>" "is_Fun m"
  shows "\<exists>m' \<in> set (var_disj M). \<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> s = m' \<cdot> \<theta> \<and> is_Fun m'"
proof -
  obtain ma da where ma: "ma\<in>set (var_disj M)" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t da" "subst_range da \<subseteq> range Var" "m = ma \<cdot> da"
    using var_disj_is_var_renaming[OF assms(1)] by moura
  have "is_Fun ma" using ma(3,4) assms(4) by fastforce
  thus ?thesis using assms(2,3) ma by (metis (full_types) subst_subst_compose wt_subst_compose) 
qed

lemma Ana_keys_subterm: "Ana t = (K,T) \<Longrightarrow> k \<in> K \<Longrightarrow> k \<sqsubset> t"
proof (induct t rule: Ana.induct)
  case (1 U)
  then obtain m where "U = [Fun pub [k], m]" "K = {k}" "T = {m}"
    by (auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)
  thus ?case using Fun_subterm_inside_params[of k crypt U] by auto
qed (auto elim!: Ana\<^sub>c\<^sub>r\<^sub>y\<^sub>p\<^sub>t.elims Ana\<^sub>s\<^sub>i\<^sub>g\<^sub>n.elims)

lemma is_TComp_var_instance_closedD:
  assumes x: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma>\<^sub>v x = TComp f T"
    and closed: "is_TComp_var_instance_closed M"
  shows "\<exists>g U. Fun g U \<in> set M \<and> \<Gamma> (Fun g U) = \<Gamma>\<^sub>v x \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
proof -
  let ?P = "\<lambda>g U v. \<Gamma> (Fun g U) = \<Gamma>\<^sub>v v \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
  let ?Q = "\<lambda>v. (\<exists>t \<in> set M. case t of (Fun g U) \<Rightarrow> ?P g U v | _ \<Rightarrow> False)"
  let ?R = "\<lambda>v. case \<Gamma>\<^sub>v v of (TComp f T) \<Rightarrow> ?Q v | _ \<Rightarrow> True"
  have "?R x" using x(1) closed unfolding is_TComp_var_instance_closed_def by blast
  hence "?Q x" using x(2) by simp
  then obtain t where t: "t \<in> set M" "case t of Var va \<Rightarrow> False | Fun g U \<Rightarrow> ?P g U x" by moura
  then obtain g U where "t = Fun g U" by (metis (no_types, lifting) is_FunE term.case_eq_if) 
  thus ?thesis using t by auto
qed

lemma is_TComp_var_instance_closedD':
  assumes "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "TComp f T \<sqsubseteq> \<Gamma>\<^sub>v x"
    and closed: "is_TComp_var_instance_closed M" "is_subterm_closed M"
    and wf: "is_wellformed_terms M"
  shows "\<exists>g U. Fun g U \<in> set M \<and> \<Gamma> (Fun g U) = TComp f T \<and> (\<forall>u \<in> set U. is_Var u) \<and> distinct U"
using assms(1,2)
proof (induction "\<Gamma>\<^sub>v x" arbitrary: x)
  case (Fun g U)
  note IH = Fun.hyps(1)
  have g: "arity g > 0" "public g" using Fun.hyps(2) fun_type_inv[of "Var x"] by simp_all
  then obtain V where V:
      "Fun g V \<in> set M" "\<Gamma> (Fun g V) = \<Gamma>\<^sub>v x" "\<forall>v \<in> set V. \<exists>x. v = Var x"
      "distinct V" "length U = length V"
    using is_TComp_var_instance_closedD[OF Fun.prems(1) Fun.hyps(2)[symmetric] closed(1)]
    by (metis Fun.hyps(2) fun_type_id_eq fun_type_length_eq is_VarE)
  hence U: "U = map \<Gamma> V" using fun_type[OF g(1)] Fun.hyps(2) by simp
  hence 1: "\<And>v. v \<in> set V \<Longrightarrow> \<Gamma> v \<in> set U" by simp
  have 2: "\<And>v. v \<in> set V \<Longrightarrow> v \<in> set M"
    using closed(2) V(1) unfolding is_subterm_closed_def
    by force
  hence 3: "\<And>y. Var y \<in> set V \<Longrightarrow> y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)"
    using fv_subset[OF 2] by force
  show ?case
  proof (cases "TComp f T = \<Gamma>\<^sub>v x")
    case True thus ?thesis using V by fastforce
  next
    case False
    then obtain u where u: "u \<in> set U" "TComp f T \<sqsubseteq> u"
      using Fun.prems(2) Fun.hyps(2) unfolding subterm_def by moura
    then obtain y where y: "Var y \<in> set V" "\<Gamma>\<^sub>v y = u" using U V(3) by auto
    show ?thesis using IH[OF _ 3[OF y(1)]] u y(2) by metis
  qed
qed simp

lemma TComp_var_instance_wt_subst_exists:
  assumes gT: "\<Gamma> (Fun g T) = TComp g (map \<Gamma> U)" "wf\<^sub>t\<^sub>r\<^sub>m (Fun g T)"
    and U: "\<forall>u \<in> set U. \<exists>y. u = Var y" "distinct U"
  shows "\<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>) \<and> Fun g T = Fun g U \<cdot> \<theta>"
proof -
  define the_i where the_i: "the_i \<equiv> \<lambda>y. THE x. x < length U \<and> U ! x = Var y"
  define \<theta> where \<theta>: "\<theta> \<equiv> \<lambda>y. if Var y \<in> set U then T ! the_i y else Var y"

  have g: "arity g > 0" using gT(1,2) fun_type_inv(1) by blast

  have UT: "length U = length T" using fun_type_length_eq gT(1) by auto

  have 1: "\<And>y. Var y \<in> set U \<Longrightarrow> the_i y < length U \<and> U ! the_i y = Var y"
    using theI'[OF distinct_Ex1[OF U(2)]] the_i by simp

  have 2: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>"
    using \<theta> 1 gT(1) fun_type[OF g] UT
    unfolding wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def
    by (metis (no_types, lifting) nth_map term.inject(2))

  have "\<forall>i<length T. U ! i \<cdot> \<theta> = T ! i"
    using \<theta> 1 U(1) UT distinct_Ex1[OF U(2)] in_set_conv_nth
    by (metis (no_types, lifting) subst_apply_term.simps(1)) 
  hence "T = map (\<lambda>t. t \<cdot> \<theta>) U" by (simp add: UT nth_equalityI)
  hence 3: "Fun g T = Fun g U \<cdot> \<theta>" by simp

  have "subst_range \<theta> \<subseteq> set T" using \<theta> 1 U(1) UT by (auto simp add: subst_domain_def)
  hence 4: "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)" using gT(2) wf_trm_param by auto

  show ?thesis by (metis 2 3 4)
qed

lemma TComp_var_instance_closed_has_Fun:
  assumes "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "t \<sqsubseteq> \<delta> x" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "\<Gamma>\<^sub>v x = TComp f T" "\<Gamma> t = TComp g T'"
    and t_fun: "is_Fun t"
    and closed: "is_TComp_var_instance_closed M" "is_subterm_closed M"
    and wf: "wf\<^sub>t\<^sub>r\<^sub>m (\<delta> x)" "is_wellformed_terms M"
  shows "\<exists>m \<in> set M. \<exists>\<theta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>) \<and> t = m \<cdot> \<theta> \<and> is_Fun m"
proof -
  obtain T'' where T'': "t = Fun g T''" using assms(5) t_fun fun_type_id_eq by blast
  have g: "arity g > 0" "public g" using assms(5) fun_type_inv[of t] by simp_all
  have "TComp g T' \<sqsubseteq> \<Gamma>\<^sub>v x" using assms(2,3,5)
    by (metis \<Gamma>.simps(1) wf(1) subtermeq_imp_subtermtypeeq wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def) 
  then obtain U where U:
      "Fun g U \<in> set M" "\<Gamma> (Fun g U) = TComp g T'" "\<forall>u\<in>set U. \<exists>y. u = Var y"
      "distinct U" "length T'' = length U"
    using is_TComp_var_instance_closedD'[OF assms(1) _ closed wf(2)]
    by (metis assms(5) T'' fun_type_id_eq fun_type_length_eq is_VarE)
  hence UT': "T' = map \<Gamma> U" using fun_type[OF g(1)] assms(5) by simp

  show ?thesis
    using TComp_var_instance_wt_subst_exists UT' T'' U(1,3,4) assms(2,5) wf(1)
    by (metis term.disc(2) wf_trm_subtermeq)
qed

lemma comp_tfr_SMP_D_aux:
  assumes hyps: "t' \<sqsubset> t" and wf: "wf\<^sub>t\<^sub>r\<^sub>m t'" and prems: "is_Fun t'"
    and IH: "(\<exists>f. t = Fun f []) \<or>
             (\<exists>m\<in>set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t = m \<cdot> \<delta> \<and> is_Fun m)"
    and closed: "is_subterm_closed M" "is_TComp_var_instance_closed M"
    and wf_M: "is_wellformed_terms M"
  shows "(\<exists>f. t' = Fun f []) \<or>
         (\<exists>m\<in>set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t' = m \<cdot> \<delta> \<and> is_Fun m)"
using hyps
proof (cases "\<exists>f. t = Fun f [] \<or> t' = Fun f []")
  case False
  note F = False
  then obtain m \<delta> where m:
      "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "t = m \<cdot> \<delta>" "is_Fun m" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)"
    using IH by moura
  obtain f T where fT: "t' = Fun f T" "arity f > 0"
    using wf F prems
    by (metis is_FunE length_greater_0_conv subtermeqI' wf\<^sub>t\<^sub>r\<^sub>m_def)
  show ?thesis
  proof (cases "\<exists>x \<in> fv m. t' \<sqsubseteq> \<delta> x")
    case True
    then obtain x where x: "x \<in> fv m" "t' \<sqsubseteq> \<delta> x" by moura
    have 1: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using m(1) x(1) by auto
    have 2: "is_Fun (\<delta> x)" using prems x(2) by auto
    have 3: "wf\<^sub>t\<^sub>r\<^sub>m (\<delta> x)" using m(5) by (simp add: wf_trm_img_subst)
    have "\<not>(\<exists>f. \<delta> x = Fun f [])" using F x(2) by auto
    hence "\<exists>f T. \<Gamma>\<^sub>v x = TComp f T" using 2 3 m(2)
      by (metis (no_types) fun_type is_FunE length_greater_0_conv subtermeqI'
                           \<Gamma>.simps(1) wf\<^sub>t\<^sub>r\<^sub>m_def wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def)
    moreover have "\<exists>f T. \<Gamma> t' = Fun f T"
      using False prems wf
      by (metis (no_types) fun_type is_FunE length_greater_0_conv subtermeqI' wf\<^sub>t\<^sub>r\<^sub>m_def)
    ultimately show ?thesis
      using TComp_var_instance_closed_has_Fun 1 x(2) m(2) prems closed 3 wf_M
      by metis
  next
    case False
    then obtain m' where m': "m' \<sqsubseteq> m" "t' = m' \<cdot> \<delta>" "is_Fun m'"
      using hyps m(3) subterm_subst_not_img_subterm
      unfolding subterm_def by blast
    hence "m' \<in> set M" using closed(1) m(1) unfolding is_subterm_closed_def by auto
    thus ?thesis using m'(2,3) m(2,4,5) fT by auto
  qed
qed auto

lemma comp_tfr_SMP_D_aux':
  assumes "t \<in> SMP (set M)"
    and closed: "is_subterm_closed M" "is_Ana_closed M" "is_TComp_var_instance_closed M"
    and wf_M: "is_wellformed_terms M"
  shows "\<forall>x \<in> fv t. \<exists>y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
using assms(1)
proof (induction t rule: SMP.induct)
  case (MP t)
  thus ?case
  proof (induction t)
    case (Var x)
    hence "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using fv_subset[OF MP.hyps] by force
    thus ?case using Var by auto
  next
    case (Fun f T)
    have "\<And>s. s \<in> set T \<Longrightarrow> s \<in> set M"
      using closed(1) Fun.prems
      unfolding is_subterm_closed_def
      by force
    thus ?case using Fun.IH by auto
  qed
next
  case (Subterm t t')
  hence "fv t' \<subseteq> fv t" using subtermeq_vars_subset unfolding subterm_def by auto
  thus ?case using Subterm.IH by auto
next
  case (Substitution t \<delta>)
  note IH = Substitution.IH
  show ?case
  proof
    fix x assume x: "x \<in> fv (t \<cdot> \<delta>)"
    then obtain y where y: "y \<in> fv t" "\<Gamma> (Var x) \<sqsubseteq> \<Gamma> (Var y)"
      using Substitution.hyps(2,3)
      by (metis subst_apply_img_var subtermeqI' subtermeq_imp_subtermtypeeq
                vars_iff_subtermeq wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t_def)
    let ?P = "\<lambda>x. \<exists>y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
    show "?P x" using y IH
    proof (induction "\<Gamma> (Var y)" arbitrary: y t)
      case (Var a)
      hence "\<Gamma> (Var x) = \<Gamma> (Var y)" by auto
      thus ?case using Var(2,4) by auto
    next
      case (Fun f T)
      obtain z where z: "z \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma> (Var z) = \<Gamma> (Var y)" using Fun.prems(1,3) by blast
      show ?case
      proof (cases "\<Gamma> (Var x) = \<Gamma> (Var y)")
        case True thus ?thesis using Fun.prems by auto
      next
        case False
        then obtain \<tau> where \<tau>: "\<tau> \<in> set T" "\<Gamma> (Var x) \<sqsubseteq> \<tau>" using Fun.prems(2) Fun.hyps(2) by auto
        then obtain U where U:
            "Fun f U \<in> set M" "\<Gamma> (Fun f U) = \<Gamma> (Var z)" "\<forall>u \<in> set U. \<exists>v. u = Var v" "distinct U"
          using is_TComp_var_instance_closedD'[OF z(1) _ closed(3,1) wf_M] Fun.hyps(2) z(2)
          by (metis \<Gamma>.simps(1) fun_type_id_eq subtermeqI' is_VarE)
        hence 1: "\<forall>x \<in> fv (Fun f U). \<exists>y\<in>fv\<^sub>s\<^sub>e\<^sub>t (set M). \<Gamma> (Var y) = \<Gamma> (Var x)"
          using closed(1) U(1) unfolding is_subterm_closed_def by force
        have "arity f > 0" using U(2) z(2) Fun.hyps(2) fun_type_inv(1) by auto 
        hence "\<Gamma> (Fun f U) = TComp f (map \<Gamma> U)" using fun_type by auto
        then obtain u where u: "Var u \<in> set U" "\<Gamma> (Var u) = \<tau>"
          using \<tau>(1) U(2,3) z(2) Fun.hyps(2) by auto
        show ?thesis
          using Fun.hyps(1)[of u "Fun f U"] u \<tau> 1
          by force
      qed
    qed
  qed
next
  case (Ana t K T k)
  have "fv k \<subseteq> fv t" using Ana_keys_fv[OF Ana.hyps(2)] Ana.hyps(3) by auto
  thus ?case using Ana.IH by auto
qed

lemma comp_tfr_SMP_D:
  assumes "t \<in> SMP (set M)" "is_Fun t"
    and wf_M: "is_wellformed_terms M"
    and closed: "is_subterm_closed M" "is_Ana_closed M" "is_TComp_var_instance_closed M"
  shows "(\<exists>f. t = Fun f []) \<or>
         (\<exists>m \<in> set M. \<exists>\<delta>. wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta> \<and> wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>) \<and> t = m \<cdot> \<delta> \<and> is_Fun m)"
using assms(1,2)
proof (induction t rule: SMP.induct)
  case (MP t)
  moreover have "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t Var" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range Var)" "t = t \<cdot> Var" by simp_all
  ultimately show ?case by blast 
next
  case (Subterm t t')
  hence t_fun: "is_Fun t" by auto
  have wf: "wf\<^sub>t\<^sub>r\<^sub>m t'"
    using Subterm.hyps(1,2) wf_M(1) SMP_wf_trm wf\<^sub>t\<^sub>r\<^sub>m_code
    unfolding is_wellformed_terms_def
    by (metis wf_trm_subterm)
  note * = Subterm.hyps(2) wf Subterm.prems Subterm.IH[OF t_fun] closed(1,3) wf_M
  show ?case by (rule comp_tfr_SMP_D_aux[OF *])
next
  case (Substitution t \<delta>)
  have wf: "wf\<^sub>t\<^sub>r\<^sub>m t"
    using Substitution.hyps(1) wf_M(1) SMP_wf_trm wf\<^sub>t\<^sub>r\<^sub>m_code
    unfolding is_wellformed_terms_def
    by metis
  hence wf': "wf\<^sub>t\<^sub>r\<^sub>m (t \<cdot> \<delta>)" using Substitution.hyps(3) wf_trm_subst by blast
  have wf_img: "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)" using Substitution.hyps(3) wf_trms_subst_img_eq by blast
  show ?case
  proof (cases "\<Gamma> t")
    case (Var a)
    hence 1: "\<Gamma> (t \<cdot> \<delta>) = TAtom a" using Substitution.hyps(2) by (metis wt_subst_trm'') 
    show ?thesis using TAtom_term_cases[OF wf' 1] is_FunE[OF Substitution.prems] by auto
  next
    case (Fun f T)
    hence 1: "\<Gamma> (t \<cdot> \<delta>) = TComp f T" using Substitution.hyps(2) by (metis wt_subst_trm'')
    have 2: "\<not>(\<exists>f. t = Fun f [])" using Fun TComp_term_cases[OF wf] by auto
    obtain T'' where T'': "t \<cdot> \<delta> = Fun f T''"
      using 1 2 fun_type_id_eq Fun Substitution.prems
      by auto
    have f: "arity f > 0" "public f" using fun_type_inv[OF 1] by metis+

    show ?thesis
    proof (cases t)
      case (Fun g U)
      then obtain m \<theta> where m: "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "t = m \<cdot> \<theta>" "is_Fun m" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)"
        using Substitution.IH Fun 2 by moura
      have "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t (\<theta> \<circ>\<^sub>s \<delta>)" "t \<cdot> \<delta> = m \<cdot> (\<theta> \<circ>\<^sub>s \<delta>)" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range (\<theta> \<circ>\<^sub>s \<delta>))"
        using wt_subst_compose[OF m(2) Substitution.hyps(2)] m(3)
              wf_trms_subst_compose[OF m(5) wf_img]
        by auto
      thus ?thesis using m(1,4) by metis
    next
      case (Var x)
      then obtain y where y: "y \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" "\<Gamma> (Var y) = \<Gamma> (Var x)"
        using comp_tfr_SMP_D_aux'[OF Substitution.hyps(1) closed wf_M] Fun
        by moura
      hence 3: "\<Gamma>\<^sub>v y = TComp f T" using Var Fun by simp
      then obtain h V where V:
          "Fun h V \<in> set M" "\<Gamma> (Fun h V) = \<Gamma>\<^sub>v y" "\<forall>u \<in> set V. \<exists>z. u = Var z" "distinct V"
        by (metis is_VarE is_TComp_var_instance_closedD[OF y(1) 3 closed(3)])
      moreover have "length T'' = length V" using 3 V(2) fun_type_length_eq 1 T'' by auto 
      ultimately have TV: "T = map \<Gamma> V"
        by (metis fun_type[OF f(1)] 3 fun_type_id_eq term.inject(2))


      obtain \<theta> where \<theta>: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<theta>)" "t \<cdot> \<delta> = Fun h V \<cdot> \<theta>"
        using TComp_var_instance_wt_subst_exists 1 3 T'' TV V(2,3,4) wf'
        by (metis fun_type_id_eq)

      have 9: "\<Gamma> (Fun h V) = \<Gamma> (\<delta> x)" using y(2) Substitution.hyps(2) V(2) 1 3 Var by auto

      show ?thesis using Var \<theta> 9 V(1) by force
    qed
  qed
next
  case (Ana t K T k)
  have 1: "is_Fun t" using Ana.hyps(2,3) by auto
  then obtain f U where U: "t = Fun f U" by moura

  have 2: "fv k \<subseteq> fv t" using Ana_keys_fv[OF Ana.hyps(2)] Ana.hyps(3) by auto

  have wf_t: "wf\<^sub>t\<^sub>r\<^sub>m t"
    using SMP_wf_trm[OF Ana.hyps(1)] wf\<^sub>t\<^sub>r\<^sub>m_code wf_M
    unfolding is_wellformed_terms_def
    by auto
  hence wf_k: "wf\<^sub>t\<^sub>r\<^sub>m k"
    using Ana_keys_wf'[OF Ana.hyps(2)] wf\<^sub>t\<^sub>r\<^sub>m_code Ana.hyps(3)
    by auto

  { assume Ana_subst': "\<And>f T \<delta> K M. Ana (Fun f T) = (K,M) \<Longrightarrow> Ana (Fun f T \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,M \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
      and Ana_const: "\<And>c. arity c = 0 \<Longrightarrow> Ana (Fun c []) = ({},{})" 
    have "U \<noteq> []"
      using Ana_const Ana.hyps(2,3) U wf_t 1 TComp_term_cases[OF wf_t]
      by (metis empty_iff fun_type is_Var_def neq0_conv prod.inject term.inject(2))
    then obtain m \<delta> where m: "m \<in> set M" "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<delta>" "wf\<^sub>t\<^sub>r\<^sub>m\<^sub>s (subst_range \<delta>)" "t = m \<cdot> \<delta>" "is_Fun m"
      using Ana.IH[OF 1] U by auto
    hence "Ana (t \<cdot> \<delta>) = (K \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,T \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)" using Ana_subst' U Ana.hyps(2) by auto
    obtain Km Tm where Ana_m: "Ana m = (Km,Tm)" by moura
    hence "Ana (m \<cdot> \<delta>) = (Km \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>,Tm \<cdot>\<^sub>s\<^sub>e\<^sub>t \<delta>)"
      using Ana_subst' U m(4) is_FunE[OF m(5)] Ana.hyps(2)
      by metis
    then obtain km where km: "km \<in> Km" "k = km \<cdot> \<delta>" using Ana.hyps(2,3) m(4) by auto
    hence km_in: "km \<in> set M" using closed(2) Ana_m m(1) unfolding is_Ana_closed_def by force
    hence ?case using m(2,3) km(2)
    proof (cases "is_Fun km")
      case False
      note F = False
      then obtain x where x: "km = Var x" by auto
      hence 3: "x \<in> fv\<^sub>s\<^sub>e\<^sub>t (set M)" using fv_subset[OF km_in] by auto
      obtain kf kT where kf: "k = Fun kf kT" using Ana.prems by auto
      show ?thesis
      proof (cases "\<exists>f. k = Fun f []")
        case False
        hence 4: "arity kf > 0" using wf_k kf TAtom_term_cases const_type by fastforce
        then obtain kT' where kT': "\<Gamma> k = TComp kf kT'" by (simp add: fun_type kf) 
        then obtain V where V:
            "Fun kf V \<in> set M" "\<Gamma> (Fun kf V) = \<Gamma>\<^sub>v x" "\<forall>u \<in> set V. \<exists>v. u = Var v"
            "distinct V" "is_Fun (Fun kf V)"
          using x m(2) km(2) is_TComp_var_instance_closedD[OF 3 _ closed(3)]
          by (metis fun_type_id_eq term.disc(2) \<Gamma>.simps(1) wt_subst_trm'' is_VarE)
        have 5: "kT' = map \<Gamma> V"
          using fun_type[OF 4] x kT' km(2) m(2) V(2)
          by (metis \<Gamma>.simps(1) term.inject(2) wt_subst_trm'')
        thus ?thesis
          using TComp_var_instance_wt_subst_exists wf_k kf 4 V(3,4) kT' V(1,5)
          by metis
      qed simp
    qed auto
  }


  have 5: "k \<sqsubset> t" using Ana_keys_subterm[OF Ana.hyps(2,3)] by auto

  note * = 5 wf_k Ana.prems Ana.IH[OF 1] closed(1,3) wf_M
  show ?case by (rule comp_tfr_SMP_D_aux[OF *])
qed

lemma tfr\<^sub>s\<^sub>e\<^sub>t_if_comp_tfr:
  assumes "comp_tfr M"
  shows "tfr\<^sub>s\<^sub>e\<^sub>t (set M)"
proof -
  have *: "is_wellformed_terms M"
          "is_subterm_closed M" "is_Ana_closed M"
          "is_TComp_var_instance_closed M"
          "is_pairwise_var_disjoint (var_disj M)"
    using assms by (metis comp_tfr_def)+

  { fix s t \<delta> assume st: "s \<in> SMP (set M) - Var`\<V>" "t \<in> SMP (set M) - Var`\<V>" "\<Gamma> s \<noteq> \<Gamma> t"
    hence **: "s \<in> SMP (set M)" "is_Fun s" "t \<in> SMP (set M)" "is_Fun t" by auto
    have "\<not>(\<exists>\<delta>. Unifier \<delta> s t)"
    proof (cases "\<exists>f. s = Fun f [] \<or> t = Fun f []")
      case True thus ?thesis using st(3) **(2,4) by auto
    next
      case False
      then obtain \<sigma> s0 \<theta> t0 where
            m: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<sigma>" "s0 \<in> set (var_disj M)" "s = s0 \<cdot> \<sigma>" "is_Fun s0"
        and n: "wt\<^sub>s\<^sub>u\<^sub>b\<^sub>s\<^sub>t \<theta>" "t0 \<in> set (var_disj M)" "t = t0 \<cdot> \<theta>" "is_Fun t0"
        using comp_tfr_SMP_D[OF **(1,2) *(1,2,3,4)]
              comp_tfr_SMP_D[OF **(3,4) *(1,2,3,4)]
              var_disj_is_var_renaming'
        by moura
      hence "\<Gamma> s = \<Gamma> s0" "\<Gamma> t = \<Gamma> t0" by (metis wt_subst_trm'')+
      hence "\<Gamma> s0 \<noteq> \<Gamma> t0" by (metis st(3))
      hence "\<not>(\<exists>\<delta>. Unifier \<delta> s0 t0)"
        by (metis assms m(2,4) n(2,4) comp_tfr_def mgu_None_is_subst_neq)
      thus ?thesis
        using vars_term_disjoint_imp_unifier *(5) m(2,3) n(2,3)
        by (metis (no_types, hide_lams) is_pairwise_var_disjoint_def subst_comp)
    qed
  } thus ?thesis by (metis tfr\<^sub>s\<^sub>e\<^sub>t_def)
qed



subsection {* TLS example: Proving type-flaw resistance *}
abbreviation \<Gamma>\<^sub>v_clientHello where
  "\<Gamma>\<^sub>v_clientHello \<equiv>
    TComp clientHello [TAtom Nonce, TAtom Nonce, TAtom Nonce, TAtom Nonce, TAtom Nonce]"

abbreviation \<Gamma>\<^sub>v_serverHello where
  "\<Gamma>\<^sub>v_serverHello \<equiv>
    TComp serverHello [TAtom Nonce, TAtom Nonce, TAtom Nonce, TAtom Nonce, TAtom Nonce]"

abbreviation \<Gamma>\<^sub>v_pub where
  "\<Gamma>\<^sub>v_pub \<equiv> TComp pub [TAtom PrivKey]"

abbreviation \<Gamma>\<^sub>v_x509 where
  "\<Gamma>\<^sub>v_x509 \<equiv> TComp x509 [TAtom Agent, \<Gamma>\<^sub>v_pub]"

abbreviation \<Gamma>\<^sub>v_sign where
  "\<Gamma>\<^sub>v_sign \<equiv> TComp sign [TAtom PrivKey, \<Gamma>\<^sub>v_x509]"

abbreviation \<Gamma>\<^sub>v_serverCert where
  "\<Gamma>\<^sub>v_serverCert \<equiv> TComp serverCert [\<Gamma>\<^sub>v_sign]"

abbreviation \<Gamma>\<^sub>v_pmsForm where
  "\<Gamma>\<^sub>v_pmsForm \<equiv> TComp pmsForm [TAtom SymKey]"

abbreviation \<Gamma>\<^sub>v_crypt where
  "\<Gamma>\<^sub>v_crypt \<equiv> TComp crypt [\<Gamma>\<^sub>v_pub, \<Gamma>\<^sub>v_pmsForm]"

abbreviation \<Gamma>\<^sub>v_clientKeyExchange where
  "\<Gamma>\<^sub>v_clientKeyExchange \<equiv>
    TComp clientKeyExchange [\<Gamma>\<^sub>v_crypt]"

abbreviation \<Gamma>\<^sub>v_HSMsgs where
  "\<Gamma>\<^sub>v_HSMsgs \<equiv> TComp concat [
    \<Gamma>\<^sub>v_clientHello,
    \<Gamma>\<^sub>v_serverHello,
    \<Gamma>\<^sub>v_serverCert,
    TAtom PubConst,
    \<Gamma>\<^sub>v_clientKeyExchange]"

(* Variables from TLS *)
abbreviation "T\<^sub>1 n \<equiv> Var (TAtom Nonce,n)"
abbreviation "T\<^sub>2 n \<equiv> Var (TAtom Nonce,n)"
abbreviation "R\<^sub>A n \<equiv> Var (TAtom Nonce,n)"
abbreviation "R\<^sub>B n \<equiv> Var (TAtom Nonce,n)"
abbreviation "S n \<equiv> Var (TAtom Nonce,n)"
abbreviation "Cipher n \<equiv> Var (TAtom Nonce,n)"
abbreviation "Comp n \<equiv> Var (TAtom Nonce,n)"
abbreviation "B n \<equiv> Var (TAtom Agent,n)"
abbreviation "Pr\<^sub>c\<^sub>a n \<equiv> Var (TAtom PrivKey,n)"
abbreviation "PMS n \<equiv> Var (TAtom SymKey,n)"
abbreviation "P\<^sub>B n \<equiv> Var (TComp pub [TAtom PrivKey],n)"
abbreviation "HSMsgs n \<equiv> Var (\<Gamma>\<^sub>v_HSMsgs,n)"

(* Variables for all composed types occurring in the type of TLS variables *)
abbreviation "V_clientHello n \<equiv> Var (\<Gamma>\<^sub>v_clientHello,n)"
abbreviation "V_serverHello n \<equiv> Var (\<Gamma>\<^sub>v_serverHello,n)"
abbreviation "V_pub n \<equiv> Var (\<Gamma>\<^sub>v_pub,n)"
abbreviation "V_x509 n \<equiv> Var (\<Gamma>\<^sub>v_x509,n)"
abbreviation "V_sign n \<equiv> Var (\<Gamma>\<^sub>v_sign,n)"
abbreviation "V_serverCert n \<equiv> Var (\<Gamma>\<^sub>v_serverCert,n)"
abbreviation "V_pmsForm n \<equiv> Var (\<Gamma>\<^sub>v_pmsForm,n)"
abbreviation "V_crypt n \<equiv> Var (\<Gamma>\<^sub>v_crypt,n)"
abbreviation "V_clientKeyExchange n \<equiv> Var (\<Gamma>\<^sub>v_clientKeyExchange,n)"


subsubsection {* Defining the over-approximation set *}
abbreviation clientHello\<^sub>t\<^sub>r\<^sub>m where
  "clientHello\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun clientHello [T\<^sub>1 0, R\<^sub>A 0, S 0, Cipher 0, Comp 0]"

abbreviation serverHello\<^sub>t\<^sub>r\<^sub>m where
  "serverHello\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun serverHello [T\<^sub>2 0, R\<^sub>B 0, S 0, Cipher 0, Comp 0]"

abbreviation serverCert\<^sub>t\<^sub>r\<^sub>m where
  "serverCert\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun serverCert [Fun sign [Pr\<^sub>c\<^sub>a 0, Fun x509 [B 0, P\<^sub>B 0]]]"

abbreviation serverHelloDone\<^sub>t\<^sub>r\<^sub>m where
  "serverHelloDone\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun serverHelloDone []"

abbreviation clientKeyExchange\<^sub>t\<^sub>r\<^sub>m where
  "clientKeyExchange\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun clientKeyExchange [Fun crypt [P\<^sub>B 0, Fun pmsForm [PMS 0]]]"

abbreviation changeCipher\<^sub>t\<^sub>r\<^sub>m where
  "changeCipher\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun changeCipher []"

abbreviation finished\<^sub>t\<^sub>r\<^sub>m where
  "finished\<^sub>t\<^sub>r\<^sub>m \<equiv> Fun finished [Fun prfun [
      Fun clientFinished [
          Fun prfun [Fun master [PMS 0, R\<^sub>A 0, R\<^sub>B 0]],
          R\<^sub>A 0, R\<^sub>B 0, Fun hash [HSMsgs 0]
      ]
  ]]"

abbreviation M\<^sub>T\<^sub>L\<^sub>S::"ex_term list" where
  "M\<^sub>T\<^sub>L\<^sub>S \<equiv> [
      clientHello\<^sub>t\<^sub>r\<^sub>m,
      serverHello\<^sub>t\<^sub>r\<^sub>m,
      serverCert\<^sub>t\<^sub>r\<^sub>m,
      serverHelloDone\<^sub>t\<^sub>r\<^sub>m,
      clientKeyExchange\<^sub>t\<^sub>r\<^sub>m,
      changeCipher\<^sub>t\<^sub>r\<^sub>m,
      finished\<^sub>t\<^sub>r\<^sub>m,

(* composed-type variable instantiations *)
Fun pub [Var (TAtom PrivKey, 0)],
Fun concat [V_clientHello 0, V_serverHello 0, V_serverCert 0,
            Var (TAtom PubConst, 0), V_clientKeyExchange 0],
Fun clientHello [Var (TAtom Nonce, 0), Var (TAtom Nonce, 1), Var (TAtom Nonce, 2),
                 Var (TAtom Nonce, 3), Var (TAtom Nonce, 4)],
Fun serverHello [Var (TAtom Nonce, 0), Var (TAtom Nonce, 1), Var (TAtom Nonce, 2),
                 Var (TAtom Nonce, 3), Var (TAtom Nonce, 4)],
Fun serverCert [V_sign 0],
Fun sign [Var (TAtom PrivKey, 0), V_x509 0],
Fun x509 [Var (TAtom Agent, 0), V_pub 0],
Fun clientKeyExchange [V_crypt 0],
Fun crypt [V_pub 0, V_pmsForm 0],
Fun pmsForm [Var (TAtom SymKey, 0)],

(* subterm closure *)
T\<^sub>1 0, R\<^sub>A 0, S 0, Cipher 0, Comp 0,
T\<^sub>2 0, R\<^sub>B 0,
Fun sign [Pr\<^sub>c\<^sub>a 0, Fun x509 [B 0, P\<^sub>B 0]],
Pr\<^sub>c\<^sub>a 0, Fun x509 [B 0, P\<^sub>B 0],
B 0, P\<^sub>B 0,
Fun crypt [P\<^sub>B 0, Fun pmsForm [PMS 0]],
Fun pmsForm [PMS 0],
PMS 0,
Fun prfun [
  Fun clientFinished [
    Fun prfun [Fun master [PMS 0, R\<^sub>A 0, R\<^sub>B 0]],
    R\<^sub>A 0, R\<^sub>B 0, Fun hash [HSMsgs 0]
  ]
],
Fun clientFinished [
  Fun prfun [Fun master [PMS 0, R\<^sub>A 0, R\<^sub>B 0]],
  R\<^sub>A 0, R\<^sub>B 0, Fun hash [HSMsgs 0]
],
Fun prfun [Fun master [PMS 0, R\<^sub>A 0, R\<^sub>B 0]],
Fun master [PMS 0, R\<^sub>A 0, R\<^sub>B 0],
Fun hash [HSMsgs 0],
HSMsgs 0,
Var (TAtom PrivKey, 0),
V_clientHello 0, V_serverHello 0, V_serverCert 0,
Var (TAtom PubConst, 0), V_clientKeyExchange 0,
Var (TAtom Nonce, 0), Var (TAtom Nonce, 1), Var (TAtom Nonce, 2),
Var (TAtom Nonce, 3), Var (TAtom Nonce, 4),
V_sign 0,
V_x509 0,
Var (TAtom Agent, 0), V_pub 0,
V_crypt 0,
V_pub 0, V_pmsForm 0,
Var (TAtom SymKey, 0)
]"


subsection {* Theorem: The TLS handshake protocol is type-flaw resistant *}
theorem "tfr\<^sub>s\<^sub>e\<^sub>t (set M\<^sub>T\<^sub>L\<^sub>S)"
proof -
  have "comp_tfr M\<^sub>T\<^sub>L\<^sub>S" by eval (* Note: eval is faster than code_simp but also potentially unsafe *)
  thus ?thesis by (rule tfr\<^sub>s\<^sub>e\<^sub>t_if_comp_tfr)
qed

end
